package ninja.thefirearchmage.games.fourtykcalculator;

import ninja.thefirearchmage.games.fourtykcalculator.utils.datastructures.Tuple;
import ninja.thefirearchmage.games.fourtykcalculator.utils.datastructures.Tuple2;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

public class AttackScenario {
    private static final int PROBABILITY_DENOMINATOR_D6 = 6;

    // Attack attributes
    private UnitAttackModifiers unitAttackModifiers;
    private List<Tuple2<Weapon, Integer>> weapons;
    private boolean unmodifiableHit;
    private boolean attackerWasStationary;
    private boolean attackerCharged;
    private boolean attackerInRapidFireRange;
    private boolean hasHitRollBonus;
    private boolean hasWoundRollBonus;
    private boolean hasLineOfSight;
    private boolean hasRerollHitsOnes;
    private boolean hasRerollHitsFailures;
    private boolean hasRerollHitsAny;
    private boolean hasRerollWoundsOnes;
    private boolean hasRerollWoundsFailures;
    private boolean hasRerollWoundsAny;
    private boolean fishCriticalHits;
    private boolean fishCriticalWounds;
    private int criticalHitValue;
    private int criticalWoundValue;

    // Defense attributes
    private DefensiveProfile defensiveProfile;
    private Optional<Reroll> normalSaveReroll;
    private boolean hasHitPenalty;
    private boolean hasWoundRollPenalty;
    private boolean isInCover;
    private boolean woundRollPenaltyIfStrengthIsHigher;
    private int damageReduction;

    // Result
    private AttackScenarioResolution results;

    public AttackScenario() {
        weapons = new ArrayList<>();
        unmodifiableHit = false;
        attackerWasStationary = false;
        attackerCharged = true;
        attackerInRapidFireRange = true;
        hasHitRollBonus = false;
        hasWoundRollBonus = false;
        hasLineOfSight = true;
        hasRerollHitsOnes = false;
        hasRerollHitsFailures = false;
        hasRerollHitsAny = false;
        hasRerollWoundsOnes = false;
        hasRerollWoundsFailures = false;
        hasRerollWoundsAny = false;
        fishCriticalHits = false;
        fishCriticalWounds = false;
        criticalHitValue = 6;
        criticalWoundValue = 6;

        defensiveProfile = null;
        normalSaveReroll = Optional.empty();
        hasHitPenalty = false;
        hasWoundRollPenalty = false;
        isInCover = false;
        woundRollPenaltyIfStrengthIsHigher = false;
        damageReduction = 0;

        results = new AttackScenarioResolution();
    }

    public AttackScenario addWeapon(Weapon weaponType, int numberOfWeaponsInUnit) {
        weapons.add(Tuple.of(weaponType, numberOfWeaponsInUnit));
        return this;
    }

    public AttackScenario withDefensiveProfile(DefensiveProfile defensiveProfile) {
        this.defensiveProfile = defensiveProfile;
        return this;
    }

    public AttackScenarioResolution resolveAttack() {
        // Filter out weapons that cannot participate in this attack
        var validWeapons = weapons.stream().filter( w -> hasLineOfSight || w._1.hasIndirectFire())
                .toList();

        // Attack step
        var totalDamage = 0d;
        var totalPreventedDamage = 0d;
        for(var weaponData : validWeapons) {
            Weapon weapon = weaponData._1;
            int amountOfWeapons = weaponData._2;

            var weaponAttackResolution = weapon.generateAttackResolution(defensiveProfile, unitAttackModifiers, amountOfWeapons );

            // Attack funnel
            var passingHits = calculatePassingHits(weapon, amountOfWeapons);
            var passingWounds = calculatePassingWounds(weapon, amountOfWeapons, passingHits);
            var normalWounds = passingWounds._1;
            var mortalWounds = passingWounds._2;

            // Defence funnel
            var savesResolution = defensiveProfile.saveWounds(weapon, normalWounds, mortalWounds);
            results.setWeaponStat(weapon, WeaponStat.DAMAGE_NORMAL, normalWounds - savesResolution.getPreventedNormalWounds());
            results.setWeaponStat(weapon, WeaponStat.DAMAGE_MORTAL, mortalWounds - savesResolution.getPreventedMortalWoundsFromFNP());
            results.setWeaponStat(weapon, WeaponStat.DAMAGE_TOTAL, normalWounds + mortalWounds - savesResolution.getPreventedWounds());

            totalDamage += normalWounds + mortalWounds - savesResolution.getPreventedWounds();
            totalPreventedDamage += savesResolution.getPreventedWounds();
        }
        results.setTotalPreventedDamage(totalDamage);
        results.setTotalPreventedDamage(totalPreventedDamage);

        return results;
    }


}
