package ninja.thefirearchmage.games.fourtykcalculator;

import lombok.Getter;
import ninja.thefirearchmage.games.fourtykcalculator.utils.datastructures.Tuple;
import ninja.thefirearchmage.games.fourtykcalculator.utils.datastructures.Tuple2;

import java.util.*;

public class Weapon {
    @Getter
    private String name;
    @Getter
    private int strength;
    @Getter
    private int ap;
    @Getter
    private int attacks;
    @Getter
    private WeaponDamage damage;
    @Getter
    private WeaponRangeType rangeType;
    @Getter
    private int hitValue;
    private boolean hasLethalHits;
    @Getter
    private Optional<Integer> sustainedHits;
    private boolean hasDevastatingWounds;
    private boolean hasHazardous;
    private List<AntiUnitEffect> antiUnitEffects;
    private boolean hasHeavy;
    private boolean hasBlast;
    private boolean hasLance;
    private boolean ignoresCover;
    private boolean hasIndirectFire;
    @Getter
    private Optional<Integer> melta;
    @Getter
    private Optional<Integer> rapidFire;
    private boolean isPsychic;

    public Weapon(String name, int strength, int ap, int attacks, int hitValue, WeaponDamage damage,
                  WeaponRangeType weaponRangeType) {
        this.name = name;
        this.strength = strength;
        this.ap = ap;
        this.attacks = attacks;
        this.hitValue = hitValue;
        this.damage = damage;
        this.rangeType = weaponRangeType;

        this.antiUnitEffects = new ArrayList<>();
        this.melta = Optional.empty();
        this.rapidFire = Optional.empty();
        this.sustainedHits = Optional.empty();

        this.hasDevastatingWounds = false;
        this.hasHazardous = false;
        this.hasHeavy = false;
        this.hasBlast = false;
        this.hasLance = false;
        this.ignoresCover = false;
        this.hasIndirectFire = false;
        this.hasLethalHits = false;
    }

    /**
     * Given a defensive profile, attack unit modifiers and the number of weapons to shoot, resolves the attack
     * portion statistics of the attack scenario. Returns the stats about the attack itself, before the wounds
     * are assigned to the target.
     */
    public WeaponAttackResolution generateAttackResolution(DefensiveProfile target, UnitAttackModifiers unitModifiers,
                                                           int weaponAmount) {
        var weaponAttackResolution = new WeaponAttackResolution();

        var passingHits = calculatePassingHits(target, unitModifiers, weaponAmount, weaponAttackResolution);
        var passingWounds = calculatePassingWounds(weaponAmount, passingHits);
        var normalWounds = passingWounds._1;
        var mortalWounds = passingWounds._2;

        return weaponAttackResolution;
    }


    private double calculatePassingHits(DefensiveProfile target, UnitAttackModifiers unitAttackModifiers,
                                        int amountOfWeapons, WeaponAttackResolution weaponAttackResolution) {
        int hitRollGoal = calculateHitRollGoal(target, unitAttackModifiers);
        int hitSuccessProbabilityNumerator = calculateProbabilityNumeratorFrom(hitRollGoal);
        int weaponAttacks = calculateWeaponAttacks(amountOfWeapons, unitAttackModifiers, weaponAttackResolution);

        // calculate sustained hits
        var sustainedHits = calculateSustainedHits(weaponAttacks, unitAttackModifiers, weaponAttackResolution);
        weaponAttackResolution.setHitsFromSustainedHits(sustainedHits);
        weaponAttackResolution.setTotalAttacks(weaponAttacks+sustainedHits);

        // Base passing hits without rerolls
        double passingHits = (weaponAttacks*amountOfWeapons) * ((double) hitSuccessProbabilityNumerator / PROBABILITY_DENOMINATOR_D6);
        results.setWeaponStat(weapon, WeaponStat.HITS_COUNT_WITHOUT_REROLL, passingHits);

        // Add hits from rerolls
        var hitRerolls = prepareHitRerolls(hitRollGoal);
        if(hitRerolls.isPresent()) {
            var rerolls = hitRerolls.get();
            var successfulHitFromRerolls = rerolls.getProbabilityNumerator() * ((double) hitSuccessProbabilityNumerator / PROBABILITY_DENOMINATOR_D6);
            results.setWeaponStat(weapon, WeaponStat.HITS_COUNT_FROM_REROLL, successfulHitFromRerolls);
            passingHits += successfulHitFromRerolls;
        }
        results.setWeaponStat(weapon, WeaponStat.HITS_TOTAL_COUNT, passingHits);

        return passingHits;
    }

    private double calculateSustainedHits(int weaponAttacks, UnitAttackModifiers unitAttackModifiers,
                                          WeaponAttackResolution weaponAttackResolution) {
        var sustainedHitsAmounts = 0d;

        var hasSustainedHits = this.sustainedHits.isPresent() ||
                (unitAttackModifiers.getSustainedHitsModifier()._1 && unitAttackModifiers.getSustainedHitsModifier()._2.appliesTo(this) );
        if(hasSustainedHits) {
            var weaponSustainedHits = this.sustainedHits.orElse(0);
            var unitModifiersSustainedHits = unitAttackModifiers.getSustainedHitsModifier()._1 ?
                    unitAttackModifiers.getSustainedHitsModifier()._3 : 0;
            var sustainedHitsValue = Math.max(weaponSustainedHits, unitModifiersSustainedHits);
            var criticalHitValue = unitAttackModifiers.getCriticalHitValue();

            double nonSustainedProbability = (double) (6 - (7 - criticalHitValue)) / 6;
            double sustainedProbability = (double) (7 - criticalHitValue) / 6;
            if(unitAttackModifiers.isFishForCriticalHits() && hasRerollAny(unitAttackModifiers.getHitRerollsModifier())) {
                sustainedProbability += nonSustainedProbability*sustainedProbability;
            }
            sustainedHitsAmounts = weaponAttacks * sustainedProbability * sustainedHitsValue;
        }

        return sustainedHitsAmounts;
    }

    /**
     * Returns a tuple of normal passing wounds and mortal wounds so that the caller can handle them differently.
     */
    private Tuple2<Double, Double> calculatePassingWounds(Weapon weapon, int amountOfWeapons, double passingHits) {
        double lethalHitsCount = calculateLethalHits(weapon, amountOfWeapons);
        var normalHits = passingHits - lethalHitsCount;
        results.setWeaponStat(weapon, WeaponStat.NON_LETHAL_HITS_COUNT, normalHits);

        int woundRollGoal = calculateWoundRollGoal(weapon);
        int woundSuccessProbabilityNumerator = calculateProbabilityNumeratorFrom(woundRollGoal);

        var passingNormalWounds = normalHits * ( (double) woundSuccessProbabilityNumerator / PROBABILITY_DENOMINATOR_D6);
        results.setWeaponStat(weapon, WeaponStat.WOUNDS_COUNT_WITHOUT_REROLL, passingNormalWounds+lethalHitsCount);
        var woundsRerolls = prepareWoundRerolls(woundRollGoal);
        if(woundsRerolls.isPresent()) {
            var rerolls = woundsRerolls.get();
            var failedWounds = normalHits * ((double) (1 - woundSuccessProbabilityNumerator) / PROBABILITY_DENOMINATOR_D6);
            var extraWoundsFromRerolls = failedWounds * rerolls.getProbabilityNumerator() / PROBABILITY_DENOMINATOR_D6;
            results.setWeaponStat(weapon, WeaponStat.WOUNDS_COUNT_FROM_REROLL, extraWoundsFromRerolls);
            passingNormalWounds += extraWoundsFromRerolls;
        }

        // Mortal wounds
        var passingMortalWounds = 0d;
        if(weapon.hasDevastatingWounds()) {
            var criticalWoundRollGoal = weapon.getBestAntiFor(defensiveProfile)
                    .orElse(criticalWoundValue);
            var criticalProbability = calculateProbabilityNumeratorFrom(criticalWoundRollGoal);
            passingMortalWounds += (normalHits*criticalProbability/PROBABILITY_DENOMINATOR_D6);
            if(fishCriticalWounds && hasRerollWoundsAny) {
                passingMortalWounds += (normalHits * (1-criticalProbability)/PROBABILITY_DENOMINATOR_D6)*criticalProbability/PROBABILITY_DENOMINATOR_D6;
            }

            results.setWeaponStat(weapon, WeaponStat.MORTAL_WOUNDS_COUNT, passingMortalWounds);
            // passing wounds was calculated from the whole, but we must substract the critical wounds from the whole
            // to be able to handle them separately, as mortal wounds replace normal wounds.
            passingNormalWounds -= passingMortalWounds;
        }

        passingNormalWounds += lethalHitsCount;
        results.setWeaponStat(weapon, WeaponStat.WOUNDS_TOTAL_COUNT, passingNormalWounds+passingMortalWounds);

        return Tuple.of(passingNormalWounds, passingMortalWounds);
    }

    private double calculateLethalHits(Weapon weapon, int amountOfWeapons) {
        var lethalHitsCount = 0d;

        if(weapon.hasLethalHits()) {
            double weaponAttacks = calculateBasicWeaponAttacks(weapon, amountOfWeapons);
            var lethalHitsProbabilityNumerator = calculateProbabilityNumeratorFrom(criticalHitValue);

            lethalHitsCount = weaponAttacks * ((double) lethalHitsProbabilityNumerator / PROBABILITY_DENOMINATOR_D6);
            results.setWeaponStat(weapon, WeaponStat.LETHAL_HITS_COUNT, lethalHitsCount);
        }

        return lethalHitsCount;
    }

    private Optional<Reroll> prepareHitRerolls(int hitRollGoal) {
        if(hasRerollHitsOnes) {
            return Optional.of(Reroll.rerollOnOnes());
        }
        if(hasRerollHitsFailures || hasRerollHitsAny) {
            return Optional.of(Reroll.rerollFailures(hitRollGoal));
        }

        return Optional.empty();
    }

    private Optional<Reroll> prepareWoundRerolls(int woundRollGoal) {
        if(hasRerollWoundsOnes) {
            return Optional.of(Reroll.rerollOnOnes());
        }
        if(hasRerollWoundsFailures || hasRerollWoundsAny) {
            return Optional.of(Reroll.rerollFailures(woundRollGoal));
        }

        return Optional.empty();
    }

    private int calculateWeaponAttacks(int amountOfWeapons, UnitAttackModifiers unitAttackModifiers,
                                          WeaponAttackResolution weaponAttackResolution) {
        double weaponAttacks = calculateBasicWeaponAttacks(amountOfWeapons, unitAttackModifiers, weaponAttackResolution);



        return weaponAttacks;
    }

    private static boolean hasRerollAny(List<Reroll> rerolls) {
        return rerolls.stream().anyMatch(r -> r.getType() == RerollType.ANY);
    }

    private double calculateBasicWeaponAttacks(int amountOfWeapons, UnitAttackModifiers unitAttackModifiers,
                                               WeaponAttackResolution weaponAttackResolution) {
        var weaponAttacks = this.attacks;

        // Rapid fire
        var hasRapidFire = this.getRapidFire().isPresent() || unitAttackModifiers.getRapidFireModifier()._1;
        if(hasRapidFire && unitAttackModifiers.isTargetInRapidFireRange()) {
            var higherRapidFireValue = Math.max(this.rapidFire.get(), unitAttackModifiers.getRapidFireModifier()._3);
            weaponAttacks += higherRapidFireValue;
            weaponAttackResolution.setAttacksFromRapidFire(higherRapidFireValue*amountOfWeapons);
        }

        // Attacks modifiers
        var attackModifier = 0;
        var attackBonus = unitAttackModifiers.getAttacksBonusModifier();
        if(attackBonus._1 && attackBonus._2.appliesTo(this)) {
            attackModifier += attackBonus._3;
        }
        var attackMalus = unitAttackModifiers.getAttacksMalusModifier();
        if(attackMalus._1 && attackMalus._2.appliesTo(this)) {
            attackModifier -= attackMalus._3;
        }
        weaponAttackResolution.setAttacksFromModifiers(attackModifier*amountOfWeapons);
        weaponAttacks += attackModifier;

        weaponAttacks *= amountOfWeapons;
        weaponAttackResolution.setTotalAttacks(weaponAttacks);

        return weaponAttacks;
    }

    private int calculateHitRollGoal(DefensiveProfile target, UnitAttackModifiers unitAttackModifiers) {
        var finalHitRollGoal = this.getHitValue();
        var hitModifier = 0;

        // Bonuses
        var unitHitBonus = unitAttackModifiers.getHitBonusModifier();
        if(unitHitBonus._1 && (unitHitBonus._2.appliesTo(this))) {
            hitModifier--;
        }
        if(this.hasHeavy && unitAttackModifiers.isUnitHasRemainedStationaryModifier()) {
            hitModifier--;
        }

        // Maluses
        var unitHitMalus = unitAttackModifiers.getHitMalusModifier();
        if(unitHitMalus._1 && !unitAttackModifiers.isUnmodifiableHitRoll() && unitHitMalus._2.appliesTo(this)) {
            hitModifier++;
        }
        if(target.isHasStealth() && target.isInStealthRange() && !unitAttackModifiers.isUnmodifiableHitRoll()) {
            hitModifier++;
        }
        if(!target.isVisible() && this.hasIndirectFire && !unitAttackModifiers.isUnmodifiableHitRoll()) {
            hitModifier++;
        }

        // Clever way to limit values to -1,0 or 1, the maximum hit roll modification
        hitModifier = Integer.compare(hitModifier,0);
        finalHitRollGoal += hitModifier;

        return finalHitRollGoal;
    }

    /**
     * Converts from a roll goal to a probability denominator of a fraction P/6. For hit and wound rolls, the die used
     * is D6.
     */
    private static int calculateProbabilityNumeratorFrom(int rollGoal) {
        // Because a roll of 1 always fails and a roll of 6 always succeeds, we modify the roll goal to be between
        // values 2-5
        var modifiedRollGoal = Math.max(rollGoal, 2);

        return 7 - modifiedRollGoal;

    }

    private int calculateWoundRollGoal(Weapon weapon) {
        var woundRollGoal = 4;
        var finalWoundPenalty = hasWoundRollPenalty ? 1 : 0;
        var toughness = defensiveProfile.getToughness();
        var strength = weapon.getStrength();

        if(strength > toughness) {
            finalWoundPenalty += woundRollPenaltyIfStrengthIsHigher ? 1 : 0;
            woundRollGoal = strength >= toughness*2 ? 2 : 3;
        } else if(strength < toughness) {
            woundRollGoal = toughness*2 >= strength ? 6 : 5;
        }

        var finalWoundBonus = hasWoundRollBonus ? 1 : 0;
        if(attackerCharged && weapon.hasLance() && weapon.getRangeType() == WeaponRangeType.MELEE) {
            finalWoundBonus++;
        }

        // Clever way to set either -1, 0 or 1, as the wound roll can only be modified, in the end by 1
        var finalWoundRollModifier = Integer.compare(finalWoundBonus - finalWoundPenalty, 0);

        // Having a -1 to wound rolls implies that the roll goal moves by 1
        woundRollGoal -= finalWoundRollModifier;

        // Now apply any anti-X if it is better than the current wound roll
        var antiWoundRoll = weapon.getBestAntiFor(defensiveProfile);
        if(antiWoundRoll.isPresent()) {
            woundRollGoal = Math.min(woundRollGoal, antiWoundRoll.get());
        }

        return woundRollGoal;
    }

    /**
     * Returns whether this weapon is anti for any of the given types
     */
    public boolean isAntiFor(Set<UnitTypeKeyword> unitTypes) {
        var antiUnitTypes = antiStats.keySet();
        return unitTypes.stream().anyMatch(antiUnitTypes::contains);
    }
    public Optional<Integer> getBestAntiFor(DefensiveProfile target) {
        var bestAnti = Optional.<Integer>empty();
        if(isAntiFor(target.getUnitTypeKeywords())) {
            var antiUnitTypes = antiStats.keySet();
            var minAnti = target.getUnitTypeKeywords().stream()
                    .filter(antiUnitTypes::contains)
                    .mapToInt(type -> antiStats.get(type))
                    .min().getAsInt();
            bestAnti = Optional.of(minAnti);
        }

        return bestAnti;
    }
}
